//! A more concrete template context type, that also includes the path to the template.
//!
//! Intended to allow strict type checking of templates.
//!
//! ## Simple Usage
//!
//! ```
//! # use serde::Serialize;
//! # use template_context::TemplateContext;
//! #
//! #[derive(Serialize)]
//! struct CommonContext {
//!     is_test: bool,
//! }
//!
//! #[derive(Serialize)]
//! struct CustomContext {
//!     name:       &'static str,
//!     address_id: &'static str,
//! }
//!
//! impl TemplateContext for CustomContext {
//!     type CommonContext = CommonContext;
//!
//!     const TEMPLATE_PATH: &'static str = "home/view";
//! }
//!
//! fn main() {
//!     let thing = CustomContext {
//!         name:       "hello",
//!         address_id: "jj",
//!     };
//!
//!     let common = CommonContext { is_test: false };
//!
//!     let output = thing.render(common);
//! }
//! ```
//!
//! ## Usage With Different Template Paths
//!
//! ```
//! # use std::marker::PhantomData;
//! #
//! # use serde::Serialize;
//! # use template_context::TemplateContext;
//! #
//! #[derive(Serialize)]
//! struct CommonContext {
//!     is_test: bool,
//! }
//!
//! // Marker types to impl `TemplateContext`
//! enum View {}
//! enum Edit {}
//!
//! #[derive(Serialize)]
//! struct CustomContext<T> {
//!     name:       &'static str,
//!     address_id: &'static str,
//!     phantom:    PhantomData<T>,
//! }
//!
//! impl TemplateContext for CustomContext<View> {
//!     type CommonContext = CommonContext;
//!
//!     const TEMPLATE_PATH: &'static str = "home/view";
//! }
//!
//! impl TemplateContext for CustomContext<Edit> {
//!     type CommonContext = CommonContext;
//!
//!     const TEMPLATE_PATH: &'static str = "home/edit";
//! }
//!
//! fn main() {
//!     let thing: CustomContext<View> = CustomContext {
//!         name:       "hello",
//!         address_id: "jj",
//!         phantom:    PhantomData,
//!     };
//!
//!     let common = CommonContext { is_test: false };
//!
//!     let output = thing.render(common);
//!
//!     // `output.name` == "home/view"
//!     println!("{:?}", output);
//! }
//! ```

use rocket_dyn_templates::{tera::Context, Template};
use serde::Serialize;

/// Trait allowing a struct to be rendered as a template.
///
/// ## Example
/// ```
/// # use serde::Serialize;
/// # use template_context::TemplateContext;
/// #
/// #[derive(Serialize)]
/// struct CommonContext {
///     is_test: bool,
/// }
///
/// #[derive(Serialize)]
/// struct CustomContext {
///     name:       &'static str,
///     address_id: &'static str,
/// }
///
/// impl TemplateContext for CustomContext {
///     type CommonContext = CommonContext;
///
///     const TEMPLATE_PATH: &'static str = "home/view";
/// }
///
/// fn main() {
///     let thing = CustomContext {
///         name:       "hello",
///         address_id: "jj",
///     };
///
///     let common = CommonContext { is_test: false };
///
///     let output = thing.render(common);
/// }
/// ```
pub trait TemplateContext: Serialize {
    type CommonContext: Serialize;
    const TEMPLATE_PATH: &'static str;

    /// Render a template
    fn render(&self, common: Self::CommonContext) -> Template {
        // Safety:
        // According to the serde docs: (https://docs.rs/serde_json/latest/serde_json/ser/fn.to_string.html):
        // "Serialization can fail if T’s implementation of Serialize decides to fail, or if T
        // contains a map with non-string keys."
        let context = Context::from_serialize(self).unwrap();
        let mut common_context = Context::from_serialize(common).unwrap();
        common_context.extend(context);

        Template::render(Self::TEMPLATE_PATH, common_context.into_json())
    }
}
